/* Account.cc - source file for the mailfilter program
 * Copyright (c) 2000  Andreas Bauer <baueran@in.tum.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <strstream>
#include <fstream>
#include <string>
#include <vector>
#include <sys/time.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <stdio.h>
#include <ctype.h>
#include <regex.h>
#include <fcntl.h>
#include "Account.hh"
#include "Preferences.hh"
#include "Feedback.hh"


using namespace std;

namespace acc {

  Account::Account(string server, string user, string password, int port, pref::Preferences* newPrefs) {
    this->server = new string(server);
    this->user = new string(user);
    this->pass = new string(password);
    this->port = port;
    prefs = newPrefs;
    mySocket = 0;
    report = new fb::Feedback(prefs->getLogfile(), prefs->getMode());
  }
  
  
  Account::~Account() {
  }
  
  
  int Account::check() {
    char cmd[32];
    int messages = 0, size = 0;
    int error = login();  // Login to the POP3 server
    
    if (error == 0) {
      // How many messages are in the mail box?
      string result = sendReceiveB("STAT\r\n");

      if (result[0] != '+')
	return(-6);      // Command STAT returned an error
      else {
	messages = atoi((new string(getWord(result, 1)))->c_str()); // Number of messages in mailbox
	char status[MAX_BYTES];
#ifdef TEN     
	sprintf(status, "mailfilter: Examining %d message(s).\n", messages);
#else
	snprintf(status, sizeof(status), "mailfilter: Examining %d message(s).\n", messages);
#endif
	report->message(status);
	
	// Store message information for each message from the mail box
	for (int i = 0; i < messages; i++) {
	  try {
	    // Get message size
#ifdef TEN     
	    sprintf(cmd, "LIST %d\r\n", i + 1);
#else
	    snprintf(cmd, sizeof(cmd), "LIST %d\r\n", i + 1);
#endif
	    result = sendReceiveB(cmd);
	    
	    if (result[0] != '+')
	      return(-8);  // Command LIST returned an error
	    else
	      size = atoi((new string(getWord(result, 2)))->c_str());
	    
	    // Get message header
#ifdef TEN     
	    sprintf(cmd, "TOP %d 0\r\n", i + 1);
#else
	    snprintf(cmd, sizeof(cmd), "TOP %d 0\r\n", i + 1);
#endif
	    result = sendReceive(cmd);

	    if (result[0] != '+')
	      return(-7);  // Command TOP returned an error
	    else {
	      // Store header content, #number of email, size of email
	      if (storeMessageHeader(result, i + 1, size) == -1)
		return(-10);
	    }
	  }
	  catch (CommunicationException) {
	    throw CommunicationException(); // Pass on exception
	  }
	}
      }
      
      
      // Now filter messages and delete spam and unwanted email
      vector<Header>::iterator curMessage = headers.begin();
      while(curMessage != headers.end()) {
	// Check size and delete
	if ( (curMessage->size > prefs->getMaxsize()) && (deleteMessage(curMessage->number) == -1) ) {
	  return(-9);
	}
	// Check filters and delete
	else {
	  vector<Line>::iterator curLine = curMessage->lines.begin();
	  
	  while(curLine != curMessage->lines.end()) {
	    regex_t compFilter;
	    vector<string>::iterator curFilter = prefs->getFilters()->begin();
	  
	    while (curFilter != prefs->getFilters()->end()) {
	      string currentLine = curLine->descr->c_str();
	      currentLine.append(": ");
	      currentLine.append(curLine->content->c_str());
	      
	      // Is it necessary to use regfree(&compFilter) in order to free memory for precompiled expressions??
	      regcomp(&compFilter, curFilter->c_str(), 0);

	      // Delete spam mail here and write report
	      if (regexec(&compFilter, currentLine.c_str(), 0, NULL, 0) == 0) {
		try {
		  if (deleteMessage(curMessage->number) != -1) {
		    report->message((string)"mailfilter: Deleted " + curMessage->sender->c_str() + (string)": " + curMessage->subject->c_str() + (string)", " + curMessage->date->c_str() + (string)".\n");
		  }		
		  else
		    return(-9);
		}
		catch (CommunicationException) {
		  throw CommunicationException(); // Pass on exception
		}
	      }
	      curFilter++;
	    }

	    curLine++;
	  }
	}
	
	curMessage++;
      }

      // Close and clean up connections
      disconnectServer();
      close(mySocket);
      return 0;
    }
    else if (error == -2)
      return -2;    // DNS lookup failed, login failed
    else if (error == -3)
      return -3;    // Could not create socket, login failed
    else if (error == -4)
      return -4;    // Password incorrect, login failed
    else if (error == -5)
      return -5;    // Socket connection failed, login failed
    else
      return(-1);   // Unknown error occured, login failed
  }
  
  
  int Account::connectSocket() {
    struct sockaddr_in socketAddress;
    struct hostent *myHost;
    
    myHost = gethostbyname(server->c_str());
    
    if (!myHost)
      return(-1);    // DNS lookup failure
    else {
      memset( (char*)&socketAddress, 0, sizeof(struct sockaddr_in) );
      socketAddress.sin_family = AF_INET;
      memcpy( &(socketAddress.sin_addr.s_addr), myHost->h_addr, myHost->h_length);
      socketAddress.sin_port=htons(port);
      
      if (connect(mySocket, (struct sockaddr*)&socketAddress, sizeof(struct sockaddr)) < 0)
	return(-2);  // Socket connection failure
    }
    
    return 0;
  }
  
  
  int Account::login() {
    int bytes = 0, error = 0;
    char buffer[MAX_BYTES], cmd[32];
    
    // Create socket...
    if ( (mySocket = socket(PF_INET, SOCK_STREAM, 0)) < 0)
      return(-3);    // Could not create socket
    else {
      // ...connect to socket
      error = connectSocket();
      
      if (error == -1)
	return(-2);  // DNS lookup failure
      else if (error == -2)
	return(-5);  // Socket connection failure
      else if (error == 0) {
	bytes = read(mySocket, buffer, sizeof(buffer));
	
	// Send username
#ifdef TEN     
	sprintf(cmd, "USER %s\r\n", user->c_str());
#else
	snprintf(cmd, sizeof(cmd), "USER %s\r\n", user->c_str());
#endif
	write(mySocket, cmd, strlen(cmd));
	bytes = read(mySocket, buffer, sizeof(buffer));
	buffer[bytes] = 0;
	
	// Send password
#ifdef TEN     

	sprintf(cmd, "PASS %s\r\n", pass->c_str());    
#else

	snprintf(cmd, sizeof(cmd), "PASS %s\r\n", pass->c_str());
#endif

	write(mySocket, cmd, strlen(cmd));
	bytes = read(mySocket, buffer, sizeof(buffer));
	buffer[bytes] = 0;
	
	// Error handling
	if (buffer[0] == '-') {
	  disconnectServer();   // I could handle the return value here, but I guess that's not necessary once the connection is dead anyway
	  
	  return(-4); // Password incorrect
	}
      }
      else
	return(-1);   // Login failed - unknown error
    }
    
    return 0;
  }
  
  
  // Disconnects from server and closes the socket
  int Account::disconnectServer() {
    string result = sendReceiveB("QUIT\r\n");
    
    if (result[0] != '+')
      return(-1);
    else {
      close(mySocket);
      return 0;
    }
  }


  // Sends a command to the socket (server) and returns its response as string object;
  // This method switches the socket stream to nonblocking mode and switches back when it's done.
  // This usually takes more time than the sendReceiveB() method, but can read more bytes.
  string Account::sendReceive(string command) {
    static const int TIME_OUT = 1;
    int flags = 0, oldFlag = 0;
    struct timeval tv;
    fd_set rfds;
    string input;
    char cmd[32], buffer[MAX_BYTES];
    int counter = 0, bytes = 0, last = 0;

    // Convert the command
    strcpy(cmd, command.c_str());
 
    // Send command to server and receive response
    write(mySocket, cmd, strlen(cmd));

    // Switch to non-blocking communication
    FD_ZERO(&rfds);
    FD_SET(mySocket + 1, &rfds);
    tv.tv_sec = TIME_OUT;
    tv.tv_usec = 0;
    flags = fcntl(mySocket, F_GETFL, 0);
    oldFlag = flags;
    fcntl(mySocket, F_SETFL, flags | O_NONBLOCK);
    
    do {
      last = bytes;
      
      // Only read if the socket is ready and sending data
      // Wait up to TIME_OUT seconds for data, then throw exception if nothing happens
      if (select(mySocket, &rfds, NULL,NULL, &tv) != -1)
	bytes = read(mySocket, buffer, sizeof(buffer));
      else
	throw CommunicationException();  // No data could be retrieved within given TIME_OUT

      if (bytes < MAX_BYTES)
	buffer[bytes] = 0;
      
      counter += bytes;
      input.append(buffer);
      memset((char*)buffer, 0, strlen(buffer));
    } while(bytes >= MAX_BYTES || (last >= MAX_BYTES && bytes < MAX_BYTES));
    
    // Reset socket communication to blocking mode and increase byte counter
    fcntl(mySocket, F_SETFL, oldFlag);
    input[counter] = 0;

    // Return the server response
    return(input);
  }


  // Sends a command to the socket (server) and returns its response as string object;
  // This method is very fast, unlike sendReceive(), but returns only up to the MTU limit number of bytes.
  // Thus this should only be used if you _know_ that the response from the server is very short!!
  string Account::sendReceiveB(string command) {
    char cmd[32], buffer[MAX_BYTES];
    int bytes = 0;
    
    // Convert the command
    strcpy(cmd, command.c_str());
 
    // Send command to server and receive response
    write(mySocket, cmd, strlen(cmd));

    // Read one time from the server, up to the MTU limit or up to MAX_BYTES
    bytes = read(mySocket, buffer, sizeof(buffer));
    buffer[bytes] = 0;
      
    // Return the server response
    return((string)buffer);
  }


  // Deletes a message on the POP3 server
  int Account::deleteMessage(int mess) {
    char cmd[32];
#ifdef TEN     
    sprintf(cmd, "DELE %d\r\n", mess);
#else
    snprintf(cmd, sizeof(cmd), "DELE %d\r\n", mess);
#endif
    string result = sendReceiveB(cmd);
    
    if (result[0] != '+')
      return(-1);
    else
      return 0;
  }

  
  // Returns the n-th word in a string,
  // e.g. getWord("My name is Harry", 1) would return "name".
  string Account::getWord(string word, int n) {
    int curSpace = 0, nextSpace = 0, wordEnd = 0;

    // Find beginning of word
    for (int i = 0; i < n; i++) {
      nextSpace = word.find(' ', curSpace);
      curSpace = nextSpace + 1;
    }
    
    // Either copy everything up to the next blank character, or up to the end of the line (-1 for the cr)
    if ( (wordEnd = word.find(' ', curSpace)) == -1 )
	wordEnd = word.length() - 1;

    return(word.substr(curSpace, wordEnd));
  }
  

  // Stores the header of a message (TOP n 0) in vector<Header> headers
  int Account::storeMessageHeader(string str, int message, int size) {
    Header curHeader;
    Line curLine;

    int descrStart = 0, descrEnd = 0;
    int contentStart = 0, contentEnd = 0;

    curHeader.number = message;
    curHeader.size = size;
    string head = cleanString(str);

    // Extract information about "Subject: ...", "From: ..." and store it in messageHeaders.section
    for (int i = 0; i < head.length(); i++) {
      // Description
      if ( isupper(head[i]) ) {
	descrStart = i;
	if ( (descrEnd = head.find(":", descrStart)) == -1)
	  return(-1); // encountered malformed e-mail header. If that happens, Mailfilter needs fixing here, obviously.
	curLine.descr = new string(head.substr(descrStart, (descrEnd - descrStart)));

	// Content
	contentStart = descrEnd + 2;
	if ( (contentEnd = head.find("\n", contentStart)) == -1 ) // we only store the 1st line (in case there's more)
	  contentEnd = head.length();
	curLine.content = new string(head.substr(contentStart, (contentEnd - contentStart)));

	// Store header information for deleting
	if (curLine.descr->find("From") == 0)
	  curHeader.sender = new string(curLine.content->c_str());
	else if (curLine.descr->find("Subject") == 0)
	  curHeader.subject = new string(curLine.content->c_str());
	else if (curLine.descr->find("Date") == 0)
	  curHeader.date = new string(curLine.content->c_str());

	// Move counter
	i = contentEnd;

	// Store the current line in the header (descr + content)
	curHeader.lines.push_back(curLine);
      }
      else { 
	if ( (i = head.find("\n", i)) == -1 )
	  i = head.length();
      }
    }
    
    // Store the header with all its lines
    headers.push_back(curHeader);

    return 0;
  }

  
  // Removes special line breaks from a string
  string Account::cleanString(string dirty) {
    for (int i = 0; i <= dirty.length(); i++) {
      if ( (dirty[i] == '\r' && dirty[i+1] == '\n') || (dirty[i] == '\n' && dirty[i+1] == '\r') ) {
	dirty.replace(i, 2, "\n");
      }
    }
    
    return(dirty);
  }

}
