/* mailfilter.cc - source file for the mailfilter program
 * Copyright (c) 2000  Andreas Bauer <baueran@in.tum.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <string>
#include <iostream>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include <stdlib.h>
#include "getopt.h"
#endif

#include "mailfilter.hh"
#include "Preferences.hh"
#include "Feedback.hh"
#include "Account.hh"

using namespace std;

int main(int argc, char *argv[]) {
  vector<pref::serverInfo> myServers;
  int error = 0, c = 0, option_index = 0, mode = VERBOSE;
  struct timeval tv;
  struct timezone tz;
  string* mailfilterrc = NULL;
  string* logfile = NULL;
  time_t now;
  string prefsFile;
  extern char *optarg;
  extern int optind, opterr, optopt;
  static struct option long_options[] = {
    {"help", 0, NULL, 'h'},
    {"verbose", 0, NULL, 'v'},
    {"silent", 0, NULL, 's'},
    {"mailfilterrc", 1, NULL, 'M'},
    {"logfile", 1, NULL, 'L'},
    {"version", 0, NULL, 'V'},
    {0, 0, 0, 0}
  };
  string version = (string)"mailfilter 0.0.1";
  string help = (string)"usage: mailfilter [options] [--logfile path] [--mailfilterrc path]\n" +
    (string)"options:\n" +
    (string)"\t--help          give this help\n" +
    (string)"\t--verbose       verbose mode\n" +
    (string)"\t--silent        suppress any warnings and messages\n" +
    (string)"\t--mailfilterrc  specify .mailfilterrc file\n" +
    (string)"\t--logfile       specify logfile location\n" +
    (string)"\t--version       display version info\n";
  
  // Scan command line parameters and options
  while ( (c = getopt_long(argc, argv, "abc:d:012", long_options, &option_index)) != -1 ) {
    switch (c) {
    case 'h':
      cout << help << endl;
      exit(0);
      break;
    case 'v':
      mode = VERBOSE;
      break;
    case 's':
      mode = SILENT;
      break;
    case 'V':
      cout << version << endl;
      exit(0);
      break;
    case 'M':
      mailfilterrc = new string(optarg);
      break;
    case 'L':
      logfile = new string(optarg);
      break;
    default:
      exit(0);
    }
  }

  // Locate .mailfilterrc if none was given via command line
  if (!mailfilterrc) {
    if (getenv("HOME"))
      prefsFile = (string)getenv("HOME") + (string)"/.mailfilterrc";
    else
      prefsFile = (string)"/home/" + (string)getpwuid(getuid())->pw_name + (string)"/.mailfilterrc";
  }

  try {
    // Set some preferences
    pref::Preferences* prefs = new pref::Preferences(prefsFile);
    prefs->setMode(mode);
    if (logfile)
      prefs->setLogfile(logfile->c_str());
    
    // Prepare the mailfilter Feedback module for error messages and general output
    fb::Feedback* report = new fb::Feedback(prefs->getLogfile(), prefs->getMode());
    myServers = prefs->getServers();
    vector<pref::serverInfo>::iterator curServer = myServers.begin();

    // Create separate accounts that can also be checked separately
    while(curServer != myServers.end()) {
      acc::Account* curAccount = new acc::Account(curServer->name, curServer->user, curServer->pass, curServer->port, prefs);
      gettimeofday(&tv, &tz);
      now = tv.tv_sec;
      report->message((string)"mailfilter: 0.0.1 querying " + (string)curServer->name.c_str() + (string)" on " + (string)ctime(&now));
      
      // Now log into account and kill all the spam
      error = curAccount->check();

      if (error == -1)
	report->message((string)"mailfilter: Error: Could not login to server " + (string)curServer->name.c_str() + (string)". Unknown error.\n");
      else if (error == -2)
	report->message((string)"mailfilter: Error: DNS lookup failure for " + (string)curServer->name.c_str() + (string)". Login failed.\n");
      else if (error == -3)
	report->message("mailfilter: Error: Could not create socket. Login failed.\n");
      else if (error == -4)
	report->message("mailfilter: Error: Authentication failed. Login canceled.\n");
      else if (error == -5)
	report->message("mailfilter: Error: Could not establish server connection.\n");
      else if (error == -6)
	report->message("mailfilter: Error: Sent STAT, but server responded with an error.\n");
      else if (error == -7)
	report->message("mailfilter: Error: Sent TOP, but server responded with an error.\n");
      else if (error == -8)
	report->message("mailfilter: Error: Sent LIST, but server responded with an error.\n");
      else if (error == -9)
	report->message("mailfilter: Error: Sent DELE, but server responded with an error. Delete failed.\n");
      else if (error == -10)
	report->message("mailfilter: Error: Encountered a malformed e-mail header which could not be handled by Mailfilter. Program aborted.\n");
      
      curServer++;
      error = 0;
    }
  }
  catch (pref::MalformedPrefsFile) {
    cout << "mailfilter: Error: Preferences file has wrong number of arguments or malformed syntax." << endl;
    return(-1);
  }
  catch (pref::IOException) {
    cout << "mailfilter: Error: Preferences file (usually /home/dummy/.mailfilterrc) could not be read." << endl;
    return(-1);
  }
  catch (fb::LogfileAccess) {
    cout << "mailfilter: Error: Could not access the logfile. Check the permissions of your logfile and look in your .mailfilterrc for the correct file path." << endl;
    return(-1);
  }
  catch (acc::CommunicationException) {
    cout << "mailfilter: Error: Communication failure, probably due to a time out." << endl;
    return(-1);
  }
  
  return 0;
}
